@@ -1,7 +1,8 @@
 #!/usr/bin/perl
 # Debian task selector, mark II.
-# Copyright 2004-2006 by Joey Hess <joeyh@debian.org>.
+# Copyright 2004-2011 by Joey Hess <joeyh@debian.org>.
 # Licensed under the GPL, version 2 or higher.
+use 5.014;
 use Locale::gettext;
 use Getopt::Long;
 use warnings;
@@ -11,11 +12,12 @@ textdomain('tasksel');
 my $debconf_helper="/usr/lib/tasksel/tasksel-debconf";
 my $testdir="/usr/lib/tasksel/tests";
 my $packagesdir="/usr/lib/tasksel/packages";
-my $descdir="/usr/share/tasksel";
-my $localdescdir="/usr/local/share/tasksel";
+my $descdir="/usr/share/tasksel/descs";
+my $localdescdir="/usr/local/share/tasksel/descs";
 my $statusfile="/var/lib/dpkg/status";
 my $infodir="/usr/lib/tasksel/info";
 my $testmode=0;
+my $taskpackageprefix="task-";
 
 sub warning {
 	print STDERR "tasksel: @_\n";
@@ -80,8 +82,6 @@ sub read_task_desc {
 		}
 		if (%data) {
 			$data{relevance}=5 unless exists $data{relevance};
-			$data{shortdesc}=$data{description}->[0];
-			$data{shortdesctrans}=dgettext("debian-tasks", $data{shortdesc});
 			push @ret, \%data;
 		}
 	}
@@ -96,50 +96,65 @@ sub all_tasks {
 	map { read_task_desc($_) } list_task_descs();
 }
 
-# Returns a list of all available packages.
-sub list_avail {
-	my @list;
+sub _info_avail {
+	my %ret = ();
 	# Might be better to use the perl apt bindings, but they are not
 	# currently in base.
 	open (AVAIL, "apt-cache dumpavail|");
 	local $_;
+	my ($package, $section, $priority);
 	while (<AVAIL>) {
 		chomp;
-		if (/^Package: (.*)/) {
-			push @list, $1;
+		if (not $_) {
+			if (defined $package && defined $priority && defined $section) {
+				$ret{$package} = {
+				       	"priority" => $priority,
+					"section" => $section,
+				};
+			}
+		}
+		elsif (/^Package: (.*)/) {
+			$package = $1;
+		}
+		elsif (/^Priority: (.*)/) {
+			$priority = $1;
+		}
+		elsif (/^Section: (.*)/) {
+			$section = $1;
 		}
 	}
 	close AVAIL;
-	return @list;
+	return %ret;
 }
 
 # Returns a list of all installed packages.
 sub list_installed {
 	my @list;
-	local $/="\n\n";
-	open (STATUS, $statusfile);
-	local $_;
-	while (<STATUS>) {
-		if (/^Status: .* installed$/m && /Package: (.*)$/m) {
+	open (LIST, q{LANG=C dpkg-query -W -f='${Package} ${Status}\n' |});
+	while (<LIST>) {
+		if (/^([^ ]+) .* installed$/m) {
 			push @list, $1;
 		}
 	}
-	close STATUS;
+	close LIST;
 	return @list;
 }
 
-my %avail_pkgs;
+my %_info_avail_cache;
+
+# Returns a hash of all available packages.
+sub info_avail {
+	my $package = shift;
+	if (!%_info_avail_cache) {
+		%_info_avail_cache = _info_avail();
+	}
+	return \%_info_avail_cache;
+}
+
 # Given a package name, checks to see if it's available. Memoised.
 sub package_avail {
-	my $package=shift;
-	
-	if (! %avail_pkgs) {
-		foreach my $pkg (list_avail()) {
-			$avail_pkgs{$pkg} = 1;
-		}
-	}
-
-	return $avail_pkgs{$package} || package_installed($package);
+	my $package = shift;
+	return info_avail()->{$package} || package_installed($package);
 }
 
 my %installed_pkgs;
@@ -192,12 +207,8 @@ sub task_installed {
 }
 
 # Given task hash, returns a list of all available packages in the task.
-# If the aptitude_tasks parameter is true, then it does not expand tasks
-# that aptitude knows about, and just returns aptitude task syntax for
-# those.
 sub task_packages {
 	my $task=shift;
-	my $aptitude_tasks=shift;
 	
 	my %list;
 
@@ -205,35 +216,23 @@ sub task_packages {
 	if (ref $task->{key}) {
 		map { $list{$_}=1 } @{$task->{key}};
 	}
-		
-	if ($task->{packages} eq 'task-fields') {
-		# task-fields method is built-in for speed and to support
-		# aptitude task definitions
-		if ($aptitude_tasks) {
-			return '~t^'.$task->{task}.'$';
-		}
-		else {
-			local $/="\n\n";
-			open (AVAIL, "apt-cache dumpavail|");
-			while (<AVAIL>) {
-				if (/^Task: (.*)/m) {
-					my @tasks=split(", ", $1);
-					if (grep { $_ eq $task->{task} } @tasks) { 
-						$list{$1}=1 if /^Package: (.*)/m;
-					}
-				}
-			}
-			close AVAIL;
-		}
+	
+	if (! defined $task->{packages}) {
+		# only key
 	}
 	elsif ($task->{packages} eq 'standard') {
-		# standard method is built in since it cannot easily be
-		# implemented externally.
-		return "~pstandard", "~prequired", "~pimportant";
-	}
-	elsif ($task->{packages} eq 'manual') {
-		# manual package selection is a special case
-		return;
+		while (my ($package, $info) = each(info_avail())) {
+			my ($priority, $section) = ($info->{priority}, $info->{section});
+			if (($priority eq 'required' ||
+			     $priority eq 'important' ||
+			     $priority eq 'standard') &&
+		            # Exclude packages in non-main and library sections
+		            $section !~ /^lib|\// &&
+			    # Exclude already installed packages
+		            !package_installed($package)) {
+				$list{$package} = 1;
+			}
+		}
 	}
 	else {
 		# external method
@@ -300,25 +299,64 @@ sub hide_enhancing_tasks {
 	return $task;
 }
 
-# Converts a list of tasks into a debconf list of their short descriptions.
-sub task_to_debconf {
-	my $field = shift;
-	join ", ", map {
-		my $desc=$_->{$field};
-		if ($desc=~/, /) {
-			warning("task ".$_->{task}." contains a comma in its short description: \"$desc\"");
+# Looks up the descriptions of a set of tasks, returning a new list
+# with the shortdesc fields filled in.
+sub getdescriptions {
+	my @tasks=@_;
+
+	# If the task has a description field in the task desc file,
+	# just use it, looking up a translation in gettext.
+	@tasks = map {
+		if (defined $_->{description}) {
+			$_->{shortdesc}=dgettext("debian-tasks", $_->{description}->[0]);
 		}
-		$desc;
-	} @_;
+		$_;
+	} @tasks;
+
+	# Otherwise, a more expensive apt-cache query is done,
+	# to use the descriptions of task packages.
+	my @todo = grep { ! defined $_->{shortdesc} } @tasks;
+	if (@todo) {
+		open(APT_CACHE, "apt-cache show ".join(" ", map { $taskpackageprefix.$_->{task} } @todo)." |") || die "apt-cache show: $!";
+		local $/="\n\n";
+		while (<APT_CACHE>) {
+			my ($name)=/^Package: $taskpackageprefix(.*)$/m;
+			my ($description)=/^Description-(?:[a-z][a-z](?:_[A-Z][A-Z])?): (.*)$/m;
+			($description)=/^Description: (.*)$/m
+				unless defined $description;
+			if (defined $name && defined $description) {
+				@tasks = map {
+					if ($_->{task} eq $name) {
+						$_->{shortdesc}=$description;
+					}
+					$_;
+				} @tasks;
+			}
+		}
+		close APT_CACHE;
+	}
+
+	return @tasks;
 }
 
-# Given a first parameter that is a debconf list of short descriptions of
-# tasks, or a dependency style list of task names, and then a list of task
-# hashes, returns a list of hashes for all the tasks in the list.
+# Converts a list of tasks into a debconf list of the task short
+# descriptions.
+sub task_to_debconf {
+	join ", ", map { my $d=$_->{shortdesc}; $d=~s/,/\\,/g; $d } getdescriptions(@_);
+}
+
+# Converts a list of tasks into a debconf list of the task names.
+sub task_to_debconf_C {
+	join ", ", map { $_->{task} } @_;
+}
+
+# Given a first parameter that is a string listing task names, and then a
+# list of task hashes, returns a list of hashes for all the tasks
+# in the list.
 sub list_to_tasks {
 	my $list=shift;
-	my %desc_to_task = map { $_->{shortdesc} => $_, $_->{task} => $_ } @_;
-	return grep { defined } map { $desc_to_task{$_} } split ", ", $list;
+	my %lookup = map { $_->{task} => $_ } @_;
+	return grep { defined } map { $lookup{$_} } split /[, ]+/, $list;
 }
 
 # Orders a list of tasks for display.
@@ -328,7 +366,7 @@ sub order_for_display {
 		              || 0 ||
 		  $a->{section} cmp $b->{section}
 		              || 0 ||
-	        $a->{shortdesc} cmp $b->{shortdesc}
+	        $a->{task} cmp $b->{task}
 	} @_;
 }
 
@@ -355,8 +393,8 @@ sub task_script {
 
 sub usage {
 	print STDERR gettext(q{Usage:
-tasksel install <task>
-tasksel remove <task>
+tasksel install <task>...
+tasksel remove <task>...
 tasksel [options]
 	-t, --test          test mode; don't really do anything
 	    --new-install   automatically install some tasks
@@ -377,82 +415,35 @@ sub getopts {
 		exit(1);
 	}
 	# Special case apt-like syntax.
-	if (@ARGV && $ARGV[0] eq "install") {
-		shift @ARGV;
-		$ret{install} = shift @ARGV;
-	}
-	if (@ARGV && $ARGV[0] eq "remove") {
-		shift @ARGV;
-		$ret{remove} = shift @ARGV;
-	}
 	if (@ARGV) {
-		usage();
-		exit 1;
+		my $cmd = shift @ARGV;
+		if ($cmd eq "install") {
+			$ret{cmd_install} = \@ARGV;
+		}
+		elsif ($cmd eq "remove") {
+			$ret{cmd_remove} = \@ARGV;
+		}
+		else {
+			usage();
+			exit 1;
+		}
 	}
 	$testmode=1 if $ret{test}; # set global
 	return %ret;
 }
 
-sub main {
-	my %options=getopts();
-	my @tasks_remove;
-	my @tasks_install;
+sub interactive {
+	my $options = shift;
+	my @tasks = @_;
 
-	# Options that output stuff and don't need a full processed list of
-	# tasks.
-	if (exists $options{"task-packages"}) {
-		my @tasks=all_tasks();
-		foreach my $taskname (@{$options{"task-packages"}}) {
-			my $task=name_to_task($taskname, @tasks);
-			if ($task) {
-				print "$_\n" foreach task_packages($task);
-			}
-		}
-		exit(0);
-	}
-	elsif ($options{"task-desc"}) {
-		my $task=name_to_task($options{"task-desc"}, all_tasks());
-		if ($task) {
-			my $extdesc=join(" ", @{$task->{description}}[1..$#{$task->{description}}]);
-			print dgettext("debian-tasks", $extdesc)."\n";
-			exit(0);
-		}
-		else {
-			exit(1);
-		}
-	}
-
-	# This is relatively expensive, get the full list of available tasks and
-	# mark them.
-	my @tasks=map { hide_enhancing_tasks($_) } map { task_test($_, $options{"new-install"}, 1, 0) }
-	          grep { task_avail($_) } all_tasks();
-	
-	if ($options{"list-tasks"}) {
-		map { $_->{_installed} = task_installed($_) } @tasks;
-		print "".($_->{_installed} ? "i" : "u")." ".$_->{task}."\t".$_->{shortdesc}."\n"
-			foreach order_for_display(grep { $_->{_display} } @tasks);
-		exit(0);
-	}
-	
-	if (! $options{"new-install"}) {
+	if (! $options->{"new-install"}) {
 		# Don't install hidden tasks if this is not a new install.
 		map { $_->{_install} = 0 } grep { $_->{_display} == 0 } @tasks;
 	}
-	if ($options{"install"}) {
-		my $task=name_to_task($options{"install"}, @tasks);
-		$task->{_install} = 1 if $task;
-	}
-	if ($options{"remove"}) {
-		my $task=name_to_task($options{"remove"}, @tasks);
-		push @tasks_remove, $task;
-	}
-	
-	# The interactive bit.
-	my $interactive=0;
+
 	my @list = order_for_display(grep { $_->{_display} == 1 } @tasks);
-	if (@list && ! $options{install} && ! $options{remove}) {
-		$interactive=1;
-		if (! $options{"new-install"}) {
+	if (@list) {
+		if (! $options->{"new-install"}) {
 			# Find tasks that are already installed.
 			map { $_->{_installed} = task_installed($_) } @list;
 			# Don't install new tasks unless manually selected.
@@ -465,16 +456,16 @@ sub main {
 			map { $_->{_installed} = 0 } @list;
 		}
 		my $question="tasksel/tasks";
-		if ($options{"new-install"}) {
+		if ($options->{"new-install"}) {
 			$question="tasksel/first";
 		}
 		my @default = grep { $_->{_display} == 1 && ($_->{_install} == 1 || $_->{_installed} == 1) } @tasks;
 		my $tmpfile=`tempfile`;
 		chomp $tmpfile;
 		my $ret=system($debconf_helper, $tmpfile,
-			task_to_debconf("shortdesc", @list),
-			task_to_debconf("shortdesctrans", @list),
-			task_to_debconf("shortdesc", @default),
+			task_to_debconf_C(@list),
+			task_to_debconf(@list),
+			task_to_debconf_C(@default),
 			$question) >> 8;
 		if ($ret == 30) {
 			exit 10; # back up
@@ -501,7 +492,7 @@ sub main {
 		}
 		foreach my $task (@list) {
 			if (! $task->{_selected} && $task->{_installed}) {
-				push @tasks_remove, $task;
+				$task->{_remove} = 1;
 			}
 		}
 	}
@@ -541,7 +532,7 @@ sub main {
 				# mark them for install.
 				if (! exists $tested{$task->{task}}) {
 					$ENV{TESTING_ENHANCER}=1;
-					task_test($task, $options{"new-install"}, 0, 1);
+					task_test($task, $options->{"new-install"}, 0, 1);
 					delete $ENV{TESTING_ENHANCER};
 					$tested{$task->{task}}=$task->{_install};
 				}
@@ -562,60 +553,87 @@ sub main {
 		}
 	}
 
-	# Add tasks to install and see if any selected task requires manual
-	# selection.
-	my $manual_selection=0;
-	foreach my $task (grep { $_->{_install} } @tasks) {
-		push @tasks_install, $task;
-		if ($task->{packages} eq 'manual') {
-			$manual_selection=1;
+}
+
+sub main {
+	my %options=getopts();
+	my @tasks_remove;
+	my @tasks_install;
+
+	# Options that output stuff and don't need a full processed list of
+	# tasks.
+	if (exists $options{"task-packages"}) {
+		my @tasks=all_tasks();
+		foreach my $taskname (@{$options{"task-packages"}}) {
+			my $task=name_to_task($taskname, @tasks);
+			if ($task) {
+				print "$_\n" foreach task_packages($task);
+			}
+		}
+		exit(0);
+	}
+	elsif ($options{"task-desc"}) {
+		my $task=name_to_task($options{"task-desc"}, all_tasks());
+		if ($task) {
+			my $extdesc=join(" ", @{$task->{description}}[1..$#{$task->{description}}]);
+			print dgettext("debian-tasks", $extdesc)."\n";
+			exit(0);
+		}
+		else {
+			exit(1);
 		}
 	}
+
+	# This is relatively expensive, get the full list of available tasks and
+	# mark them.
+	my @tasks=map { hide_enhancing_tasks($_) } map { task_test($_, $options{"new-install"}, 1, 0) }
+	          grep { task_avail($_) } all_tasks();
 	
-	my @aptitude;
-	if ($manual_selection) {
-		# Manaul selection and task installs, as best
-		# aptitude can do it currently. Disables use of
-		# debconf-apt-progress.
-		@aptitude="aptitude";
+	if ($options{"list-tasks"}) {
+		map { $_->{_installed} = task_installed($_) } @tasks;
+		@tasks=getdescriptions(@tasks);
+		print "".($_->{_installed} ? "i" : "u")." ".$_->{task}."\t".$_->{shortdesc}."\n"
+			foreach order_for_display(grep { $_->{_display} } @tasks);
+		exit(0);
 	}
-	elsif (-x "/usr/bin/debconf-apt-progress") {
-		@aptitude="debconf-apt-progress";
-		push @aptitude, split(' ', $options{'debconf-apt-progress'})
-			if exists $options{'debconf-apt-progress'};
-		push @aptitude, qw{-- aptitude -q};
+	
+	if ($options{cmd_install}) {
+		@tasks_install = map { name_to_task($_, @tasks) } @{$options{cmd_install}};
+	}
+	elsif ($options{cmd_remove}) {
+		@tasks_remove = map { name_to_task($_, @tasks) } @{$options{cmd_remove}};
 	}
 	else {
-		@aptitude="aptitude";
+		interactive(\%options, @tasks);
+
+		# Add tasks to install
+		@tasks_install = grep { $_->{_install} } @tasks;
+		# Add tasks to remove
+		@tasks_remove = grep { $_->{_remove} } @tasks;
 	}
-	
+
+	my @cmd;
+	if (-x "/usr/bin/debconf-apt-progress") {
+		@cmd = "debconf-apt-progress";
+		push @cmd, split(' ', $options{'debconf-apt-progress'})
+			if exists $options{'debconf-apt-progress'};
+		push @cmd, "--";
+	}
+	push @cmd, qw{apt-get -q -y -o APT::Install-Recommends=true -o APT::Get::AutomaticRemove=true install};
+
 	# And finally, act on selected tasks.
-	if (@tasks_install || @tasks_remove || $manual_selection) {
-		my @args;
+	if (@tasks_install || @tasks_remove) {
 		foreach my $task (@tasks_remove) {
-			push @args, map { "$_-" } task_packages($task, 0);
+			push @cmd, map { "$_-" } task_packages($task);
 			task_script($task->{task}, "prerm");
 		}
 		foreach my $task (@tasks_install) {
-			push @args, task_packages($task, 1);
+			push @cmd, task_packages($task);
 			task_script($task->{task}, "preinst");
 		}
-		# If the user selected no other tasks and manual package
-		# selection, run aptitude w/o the --visual-preview parameter.
-		if (! @args && $manual_selection) {
-			my $ret=run("aptitude");
-			if ($ret != 0) {
-				error gettext("aptitude failed")." ($ret)";
-			}
-		}
-		else {
-			if ($manual_selection) {
-				unshift @args, "--visual-preview";
-			}
-			my $ret=run(@aptitude, "-y", "install", @args);
-			if ($ret != 0) {
-				error gettext("aptitude failed")." ($ret)";
-			}
+		my $ret=run(@cmd);
+		if ($ret != 0) {
+			error gettext("apt-get failed")." ($ret)";
 		}
 		foreach my $task (@tasks_remove) {
 			task_script($task->{task}, "postrm");
